/* ------------------------------------------------------------------------- */
/*
 *  argument.h
 *
 *  Copyright (c) 2004 - 2007 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Tue 16 Jan 2007 05:09:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_ARGUMENT_H
#define CLX_ARGUMENT_H

#include <string>
#include <vector>
#include <map>
#include "lexical_cast.h"
#include "tokenizer.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	/*!
	 *  basic_argument_container
	 *
	 *  BNF description of the argument is as follow:
	 *  <program> [<nokey_param>]* (['-'<key>] | ['-'+<key> [<param>]*])*
	 *
	 *  The basic_argument_container class has two containers. One is
	 *  std::map<key, param>, and the other is std::vector<nokey_param>.
	 *  This class provides the interface that the upper class can control
	 *  these two containers as one container.
	 */
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_argument_container {
	public:
		typedef CharT char_type;
		typedef unsigned int size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef std::map<string_type, std::vector<string_type> > container;
		typedef typename container::key_type key_type;
		typedef typename container::mapped_type mapped_type;
		typedef typename container::value_type value_type;
		typedef typename container::iterator iterator;
		typedef typename container::const_iterator const_iterator;
		
		// constructor and destructor
		basic_argument_container() : v_(), head_() {}
		
		virtual ~basic_argument_container() {}
		
		// operator
		bool insert(const key_type& x, const string_type& s) {
			// if the argument is <nokey_param>, x is empty.
			if (x.empty()) {
				head_.push_back(s);
				return true;
			}
			
			iterator pos = v_.find(x);
			if (pos == v_.end()) {
				mapped_type v;
				v.push_back(s);
				std::pair<iterator, bool> r = v_.insert(value_type(x, v));
				return r.second;
			}
			if (!s.empty()) pos->second.push_back(s);
			return true;
		}
		
		// access
		mapped_type operator[](const key_type& x) {
			if (x.empty()) return head_;
			else return v_[x];
		}
		
		const_iterator begin() const { return v_.begin(); }
		const_iterator end() const { return v_.end(); }
		const_iterator find(const key_type& x) const { return v_.find(x); }
		
	private:
		container v_;
		mapped_type head_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_argument
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_argument {
	public:
		typedef CharT char_type;
		typedef int int_type;
		typedef unsigned int size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef basic_argument_container<CharT, Traits> container;
		typedef typename container::key_type key_type;
		typedef typename container::mapped_type mapped_type;
		typedef typename container::value_type value_type;
		typedef mapped_type param_list;
		
		// constructor and destructor
		explicit basic_argument(char_type c = '-') :
			identifier_(c), program_(), v_() {}
		
		explicit basic_argument(int_type argc, char_type* argv[], char_type c = '-') :
			identifier_(c), program_(), v_() {
			this->scan(argc, argv);
		}
		
		virtual ~basic_argument() {}
		
		// operator
		basic_argument& scan(int_type argc, char_type* argv[]) {
			program_ = argv[0];
			string_type key;
			bool eoh = false; // End of head parameter
			for (int_type i = 1; i < argc; i++) {
				string_type tmp(argv[i]);
				if (tmp.at(0) == identifier_) {
					eoh = true;
					if (tmp.at(1) != identifier_ && tmp.size() > 2) {
						for (size_type j = 1; j < tmp.size(); j++) {
							key = tmp.at(j);
							v_.insert(key, "");
						}
						key = "";
					}
					else {
						key = tmp.substr(tmp.find_first_not_of(identifier_));
						if (i + 1 == argc || argv[i+1][0] == identifier_) v_.insert(key, "");
					}
				}
				else {
					if (eoh && key.empty()) break;
					v_.insert(key, tmp);
				}
			}
			return *this;
		}
		
		bool is_exist(const key_type& x) {
			mapped_type v;
			this->assign(x, v);
			
			if (v.empty()) return false;
			else return true;
		}
		
		// access
		string_type program() { return program_; }
		mapped_type operator[](const key_type& x) { return v_[x]; }
		
		basic_argument& operator()(mapped_type& dest) {
			return this->assign(dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(size_type i, ValueT& dest) {
			return this->assign(i, dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(ValueT& dest) {
			return this->assign(0, dest);
		}
		
		basic_argument& operator()(const char_type* key, param_list& dest) {
			return this->assign(key, dest);
		}
		
		basic_argument& operator()(const string_type& key, param_list& dest) {
			return this->assign(key, dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(const char_type* key, size_type i, ValueT& dest) {
			return this->assign(key, i, dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(const string_type& key, size_type i, ValueT& dest) {
			return this->assign(key, i, dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(const char_type* key, ValueT& dest) {
			return this->assign(key, 0, dest);
		}
		
		template <class ValueT>
		basic_argument& operator()(const string_type& key, ValueT& dest) {
			return this->assign(key, 0, dest);
		}
		
	private:
		typedef basic_tokenizer<escape_separator<CharT, Traits>, std::vector<string_type> > functor;
		
		char_type identifier_;
		string_type program_;
		container v_;
		
		basic_argument& assign(param_list& dest) {
			dest = v_[""];
			return *this;
		}
		
		template <class ValueT>
		basic_argument& assign(size_type i, ValueT& dest) {
			mapped_type v;
			this->assign(v);
			if (i < v.size()) dest = lexical_cast<ValueT>(v[i]);
			return *this;
		}
		
		basic_argument& assign(const char_type* key, param_list& dest) {
			functor f(key);
			for (size_type pos = 0; pos < f.size(); pos++) {
				if (v_.find(f.at(pos)) != v_.end()) {
					dest = v_[f.at(pos)];
					break;
				}
			}
			return *this;
		}
		
		basic_argument& assign(const string_type& key, param_list& dest) {
			return this->assign(key.c_str(), dest);
		}
		
		template <class ValueT>
		basic_argument& assign(const char_type* key, size_type i, ValueT& dest) {
			mapped_type v;
			this->assign(key, v);
			if (!v.empty() && !v.at(0).empty() && i < v.size()) {
				dest = lexical_cast<ValueT>(v[i]);
			}
			return *this;
		}
		
		template <class ValueT>
		basic_argument& assign(const string_type& key, size_type i, ValueT& dest) {
			return this->assign(key.c_str(), i, dest);
		}
	};
	
	typedef basic_argument<char> argument;
}

#endif // CLX_ARGUMENT_H
