/* ------------------------------------------------------------------------- */
/*
 *  date_time.h
 *
 *  Copyright (c) 2004 - 2007 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Tue 20 Mar 2007 17:17:42 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_DATE_TIME_H
#define CLX_DATE_TIME_H

#include <ctime>
#include <sys/time.h>

namespace clx {
	/* --------------------------------------------------------------------- */
	//  date
	/* --------------------------------------------------------------------- */
	class date {
	public:
		typedef unsigned int value_type;
		typedef std::tm tm_type;
		
		// constructor and destructor
		date() { this->now(); }
		explicit date(const tm_type& t) { this->assign(t); }
		explicit date(const std::time_t& t) { this->assign(t); }
		
		explicit date(const value_type& y, const value_type& m, const value_type& d,
			const value_type& h, const value_type& mm, const value_type& s) {
			this->assign(y, m, d, h, mm, s);
		}
		
		date& operator=(const date& t) { return this->assign(t); }
		date& operator=(const tm_type& t) { return this->assign(t); }
		date& operator=(const std::time_t& t) { return this->assign(t); }
		
		// accessor
		value_type year() const { return date_.tm_year + 1900; }
		value_type mon() const { return date_.tm_mon + 1; }
		value_type day() const { return date_.tm_mday; }
		value_type hour() const { return date_.tm_hour; }
		value_type min() const { return date_.tm_min; }
		value_type sec() const { return date_.tm_sec; }
		value_type totaldays() const { return date_.tm_yday; }
		
		// get information as the traditional std::tm type
		const tm_type& reference() const { return date_; }
		const tm_type* data() const { return &date_; }
		
		std::time_t to_native() const {
			tm_type tmp = date_;
			return std::mktime(&tmp);
		}
		
		// operator
		date& now() {
			std::time_t t = std::time(NULL);
			return this->assign(t);
		}
		
		date& assign(const date& t) {
			return this->assign(t.date_);
		}
		
		date& assign(const tm_type& t) {
			date_ = t;
			return *this;
		}
		
		date& assign(const std::time_t& t) {
			struct tm tval;
			::localtime_r(&t, &tval);
			return this->assign(tval);
		}
		
		date& assign(const value_type& y, const value_type& m, const value_type& d,
			const value_type& h, const value_type& mm, const value_type& s) {
			date_.tm_year = y - 1900;
			date_.tm_mon  = m - 1;
			date_.tm_mday = d;
			date_.tm_hour = h;
			date_.tm_min  = mm;
			date_.tm_sec  = s;
			
			std::time_t tmp = this->to_native();
			return this->assign(tmp);
		}
		
		double difference(const date& x) const {
			return std::difftime(this->to_native(), x.to_native());
		}
		
		double difference(const std::time_t& x) const {
			return std::difftime(this->to_native(), x);
		}
		
		double difference(const std::tm& x) const {
			std::time_t tmp = std::mktime(const_cast<std::tm*>(&x));
			return std::difftime(this->to_native(), tmp);
		}
		
		// comparison operator
		friend bool operator<(const date& x, const date& y) {
			if (std::difftime(y.to_native(), x.to_native()) > 0.01) return true;
			return false;
		}
		
		friend bool operator>(const date& x, const date& y) {
			if (std::difftime(x.to_native(), y.to_native()) > 0.01) return true;
			return false;
		}
		
		friend bool operator<=(const date& x, const date& y) {
			return !(x > y);
		}
		
		friend bool operator>=(const date& x, const date& y) {
			return !(x < y);
		}
		
		friend bool operator==(const date& x, const date& y) {
			if (!(x < y) && !(x > y)) return true;
			return false;
		}
		
		friend bool operator!=(const date& x, const date& y) {
			return !(x == y);
		}
		
	private:
		tm_type date_;
	};
	
	/* --------------------------------------------------------------------- */
	//  time
	/* --------------------------------------------------------------------- */
	class time {
	public:
		typedef unsigned long value_type;
		typedef timeval timeval_type;
		
		static const value_type microsec = 1000000;
		
		// constructor and destructor
		time() { this->now(); }
		explicit time(const timeval_type& t) { this->assign(t); }
		
		// accessor
		value_type hour() const { return sec_ / (60 * 60); }
		value_type min() const { return (sec_ % (60 * 60)) / 60; }
		value_type sec() const { return sec_ % 60; }
		value_type usec() const { return usec_; }
		double totalsec() const { return sec_ + usec_ / static_cast<double>(microsec); }
		
		time& operator=(const time& t) { return this->assign(t); }
		time& operator=(const timeval_type& t) { return this->assign(t); }
		
		// operator
		time& now() {
			timeval_type tmp;
			::gettimeofday(&tmp, NULL);
			return this->time_of_day(tmp);
		}
		
		time& assign(const time& t) {
			sec_  = t.sec();
			usec_ = t.usec();
			return *this;
		}
		
		time& assign(const timeval_type& t) {
			return this->time_of_day(t);
		}
		
		double difference(const time& t) const {
			if (*this < t) return (-t.difference(*this));
			return this->totalsec() - t.totalsec();
		}
		
		double difference(const timeval_type& t) const {
			time tmp(t);
			return this->difference(t);
		}
		
		// comparison operator
		friend bool operator<(const time& x, const time& y) {
			if (x.totalsec() < y.totalsec()) return true;
			return false;
		}
		
		friend bool operator>(const time& x, const time& y) {
			if (x.totalsec() > y.totalsec()) return true;
			return false;
		}
		
		friend bool operator<=(const time& x, const time& y) {
			return !(x > y);
		}
		
		friend bool operator>=(const time& x, const time& y) {
			return !(x < y);
		}
		
		friend bool operator==(const time& x, const time& y) {
			if (!(x < y) && !(x > y)) return true;
			return false;
		}
		
		friend bool operator!=(const time& x, const time& y) {
			return !(x == y);
		}
		
	private:
		value_type sec_;
		value_type usec_;
		
		time& time_of_day(const timeval_type& t) {
			clx::date today;
			today.assign(today.year(), today.mon(), today.day(), 0, 0, 0);
			sec_  = t.tv_sec - today.to_native();
			usec_ = t.tv_usec;
			return *this;
		}
	};

}

#endif // CLX_DATE_TIME_H
