/* ------------------------------------------------------------------------- */
/*
 *  ioftime.h
 *
 *  Copyright (c) 2004 - 2007 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Sat 17 Mar 2007 09:47:38 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_IOFTIME_H
#define CLX_IOFTIME_H

#include <ctime>
#include <string>
#include <stdexcept>
#include <ostream>
#include "time.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_iftime
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_iftime {
	public:
		typedef CharT char_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef date date_type;
		typedef typename date_type::tm_type tm_type;
		
		// constructor and destructor
		basic_iftime() : date_() {}
		
		explicit basic_iftime(const char_type* s, const char_type* fmt) :
			date_() { this->assign(s, fmt); }
		
		explicit basic_iftime(const string_type& s, const string_type& fmt) :
			date_() { this->assign(s, fmt); }
		
		virtual ~basic_iftime() {}
		
		// operator
		const date_type& assign(const char_type* s, const char_type* fmt) {
			tm_type tmp;
			char_type* p = ::strptime(s, fmt, &tmp);
			if (*p != '\0') throw std::runtime_error("string does not match the format");
			date_ = tmp;
			return date_;
		}
		
		const date_type& assign(const string_type& s, const string_type& fmt) {
			return this->assign(s.c_str(), fmt.c_str());
		}
		
		// accessor
		const date_type& data() const { return date_; }
		
	private:
		date_type date_;
		
		// invalid copy constructor and self-assign operator
		explicit basic_iftime(const basic_iftime& cp) {
			throw std::runtime_error("Operation is not permitted");
		}
		
		basic_iftime& operator=(const basic_iftime& cp) {
			throw std::runtime_error("Operation is not permitted");
		}
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_oftime
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>,
		int BufferSize = 1024
	>
	class basic_oftime {
	public:
		typedef CharT char_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef date date_type;
		typedef typename date_type::tm_type tm_type;
		
		// constructor and destructor
		basic_oftime() : date_(), str_() {}
		
		template <class Date>
		explicit basic_oftime(const Date& t) : date_(t), str_() {}
		
		explicit basic_oftime(const char_type* fmt) :
			date_(), str_() {
			this->assign(fmt);
		}
		
		template <class Date>
		explicit basic_oftime(const char_type* fmt, const Date& t) :
			date_(t), str_() {
			this->assign(fmt);
		}
		
		explicit basic_oftime(const string_type& fmt) :
			date_(), str_() {
			this->assign(fmt);
		}
		
		template <class Date>
		explicit basic_oftime(const string_type& fmt, const Date& t) :
			date_(t), str_() {
			this->assign(fmt);
		}
		
		template <class Date>
		basic_oftime& operator=(const Date& t) {
			date_ = t;
			return *this;
		}
		
		virtual ~basic_oftime() {}
		
		// operator
		virtual void reset() {
			date_.now();
			str_.clear();
		}
		
		const string_type& assign(const char_type* fmt) {
			char_type tmp[BufferSize];
			int n = std::strftime(tmp, sizeof(tmp), fmt, date_.data());
			if (n == 0) throw std::runtime_error("temporary buffer is overflow");
			str_ = tmp;
			return str_;
		}
		
		template <class Date>
		const string_type& assign(const char_type* fmt, const Date& t) {
			date_ = t;
			return this->assign(fmt);
		}
		
		const string_type& assign(const string_type& fmt) {
			return this->assign(fmt.c_str());
		}
		
		template <class Date>
		const string_type& assign(const string_type& fmt, const Date& t) {
			date_ = t;
			return this->assign(fmt);
		}
		
		// accessor
		const string_type& str() const { return str_; }
		
	private:
		date_type date_;
		string_type str_;
		
		// invalid copy constructor and self-assign operator
		explicit basic_oftime(const basic_oftime& t) {
			throw std::runtime_error("Operation is not permitted");
		}
		
		basic_oftime& operator=(const basic_oftime& t) {
			throw std::runtime_error("Operation is not permitted");
		}
	};
	
	typedef basic_iftime<char> iftime;
	typedef basic_oftime<char> oftime;
}

#endif // CXL_IOFTIME_H
