/* ------------------------------------------------------------------------- */
/*
 *  lexical_cast.h
 *
 *  Copyright (c) 2004 - 2006 "clown", All rights reserved.
 *  
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *  
 *  Distributed under the Boost Software License, Version 1.0. (See
 *  accompanying file LICENSE_1_0.txt or copy at
 *  http://www.boost.org/LICENSE_1_0.txt)
 *
 *  Last-modified: Fri 09 Jun 2006 02:06:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_LEXICAL_CAST_H
#define CLX_LEXICAL_CAST_H

#include <iostream>
#include <sstream>
#include <string>

namespace clx {
	/* --------------------------------------------------------------------- */
	//  bad_lexical_cast
	/* --------------------------------------------------------------------- */
	class bad_lexical_cast : public std::bad_cast {
	public:
		bad_lexical_cast() : source_(&typeid(void)), target_(&typeid(void)) {}
		explicit bad_lexical_cast(const std::type_info& source, const std::type_info& target) :
			source_(&source), target_(&target) {}
		virtual ~bad_lexical_cast() throw() {}
		
		const std::type_info& source_type() const { return *source_; }
		const std::type_info& target_type() const { return *target_; }
		
		virtual const char* what() const throw() {
			return "bad lexical cast";
		}
		
	private:
		const std::type_info* source_;
		const std::type_info* target_;
	};
	
	/* --------------------------------------------------------------------- */
	//  cast_stream
	/* --------------------------------------------------------------------- */
	template <
		class Type,
		class Source
	>
	class cast_stream {
	public:
		typedef Source src_type;
		typedef Type dest_type;
		typedef std::string string_type;
		typedef std::stringstream internal_stream;
		
		// constructor and destructor
		cast_stream() : ss_() {}
		
		virtual ~cast_stream() {}
		
		// operator
		bool operator<<(const src_type& src) {
			return !(ss_ << src).fail();
		}
		
		template <class ValueT>
		bool operator>>(ValueT& dest) {
			return ss_ >> dest;
		}
		
		// speciallized for string_type;
		bool operator>>(string_type& dest) {
			dest = ss_.str();
			return true;
		}
	
	private:
		internal_stream ss_;
	};
	
	/* --------------------------------------------------------------------- */
	/*!
	 *  lexical_cast
	 *
	 *  The lexical_cast class is a sub-set class of boost::lexical_cast.
	 *  Scale back the function of wide string support.
	 */
	/* --------------------------------------------------------------------- */
	template <class Type, class Source>
	Type lexical_cast(const Source& src) {
		cast_stream<Type, Source> ss;
		Type dest;
		
		if(!(ss << src && ss >> dest)) {
			throw(bad_lexical_cast(typeid(Source), typeid(Type)));
		}
		return dest;
	};
	
	/* --------------------------------------------------------------------- */
	//  specialized lexical_cast
	/* --------------------------------------------------------------------- */
	template <> char lexical_cast(const char& src) { return src; }
	template <> int lexical_cast(const int& src) { return src; }
	template <> long lexical_cast(const long& src) { return src; }
	template <> unsigned char lexical_cast(const unsigned char& src) { return src; }
	template <> unsigned int lexical_cast(const unsigned int& src) { return src; }
	template <> unsigned long lexical_cast(const unsigned long& src) { return src; }
	template <> double lexical_cast(const double& src) { return src; }
	template <> std::string lexical_cast(const std::string& src) { return src; }
}

#endif // CLX_LEXICAL_CAST_H
