/* ------------------------------------------------------------------------- */
/*
 *  sockaddress.h
 *
 *  Copyright (c) 2004 - 2007 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Fri 16 Mar 2007 17:52:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKADDRESS_H
#define CLX_SOCKADDRESS_H

#include <cstdlib>
#include <cstring>
#include <string>
#include <sstream>
#include <exception>
#include "inet_address.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  sockaddress_error
	/* --------------------------------------------------------------------- */
	class sockaddress_error : public std::exception {
	public:
		typedef std::string string_type;
		typedef std::stringstream sstream_type;
		
		sockaddress_error() : std::exception(), hostname_() {}
		
		explicit sockaddress_error(const std::string& host) :
			std::exception(), hostname_(host) {}
		
		explicit sockaddress_error(const char* host) :
			std::exception(), hostname_(host) {}
		
		virtual ~sockaddress_error() throw() {}
		
		virtual const char* what() {
			sstream_type tmp;
			tmp << this->hostname() << ": Non-existent address";
			return tmp.str().c_str();
		}
		
		const std::string& hostname() const { return hostname_; }
	
	private:
		std::string hostname_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_sockaddress
	/* --------------------------------------------------------------------- */
	template <int Family = AF_INET>
	class basic_sockaddress {
	public:
		typedef char char_type;
		typedef unsigned int size_type;
		typedef std::string string_type;
		typedef sockaddr_in inet_type;
		typedef sockaddr base_type;
		
		// constructor and destructor
		basic_sockaddress() : addr_() {}
		
		explicit basic_sockaddress(const inet_type& addr) : addr_(addr) {}
		
		explicit basic_sockaddress(int port) : addr_() {
			this->assign(port);
		}
		
		explicit basic_sockaddress(const char_type* host, int port) : addr_() {
			this->assign(host, port);
		}
		
		explicit basic_sockaddress(const string_type& host, int port) : addr_() {
			this->assign(host, port);
		}
		
		virtual ~basic_sockaddress() {}
		
		basic_sockaddress& operator=(const inet_type& addr) {
			return this->assign(addr);
		}
		
		// operator
		basic_sockaddress& assign(const inet_type& addr) {
			addr_ = addr;
			return *this;
		}
		
		basic_sockaddress& assign(int port) {
			std::memset((char*)&addr_, 0, sizeof(addr_));
			addr_.sin_family = Family;
			addr_.sin_port = htons(port);
#ifdef CLX_WIN32
			addr_.sin_addr.S_un.S_addr = ::htonl(INADDR_ANY);
#else
			addr_.sin_addr.s_addr = ::htonl(INADDR_ANY);
#endif // CLX_WIN32
			return *this;
		}
		
		basic_sockaddress& assign(const char_type* host, int port) {
			inet_address::address_type tmp;
			if (!::inet_aton(host, &tmp)) {
				inet_address inet;
				inet.lookup(host);
				if (inet.empty()) throw sockaddress_error(host);
				tmp = inet.at(0);
			}
			
			std::memset(&addr_, 0, sizeof(addr_));
			addr_.sin_family = Family;
			addr_.sin_port = htons(port);
			std::memcpy((char*)&addr_.sin_addr, &tmp, sizeof(tmp));
			return *this;
		}
		
		basic_sockaddress& assign(const string_type& host, int port) {
			return this->assign(host.c_str(), port);
		}
		
		// get information about socket address
		size_type size() const { return sizeof(addr_); }
		const inet_type& reference() const { return addr_; }
		const inet_type* data() const { return &addr_; }
		const base_type* base_data() const { return (base_type*)this->data(); }
		
	private:
		inet_type addr_;
	};
	
	typedef basic_sockaddress<> sockaddress;
}

#endif // CLX_SOCKADDRESS_H
